@@ -90,6 +90,7 @@ save_state_info lmt_save_state = {
         .ptr       = 0,
         .initial   = memory_data_unset,
         .offset    = 0,
+        .extra     = 0, 
     },
     .current_level    = 0,
     .current_group    = 0,
@@ -194,21 +195,83 @@ void tex_initialize_undefined_cs(void)
     set_eq_level(undefined_control_sequence, level_zero);
 }
 
-void tex_dump_equivalents_mem(dumpstream f)
+/*tex
+    Dump regions 1 to 4 of |eqtb|, the table of equivalents: glue muglue toks boxes. The table of
+    equivalents usually contains repeated information, so we dump it in compressed form: The
+    sequence of $n + 2$ values $(n, x_1, \ldots, x_n, m)$ in the format file represents $n+m$
+    consecutive entries of |eqtb|, with |m| extra copies of $x_n$, namely $(x_1, \ldots, x_n, x_n,
+    \ldots, x_n)$.
+
+    Using a different threshold saves some. In fact we have either undefined or something set
+    in the main area. Packing the sizes in one byte also saves some.
+*/
+
+# define pack_them 1
+
+static void tex_dump_equivalents_mem_hash(dumpstream f)
 {
-    /*tex
-        Dump regions 1 to 4 of |eqtb|, the table of equivalents: glue muglue toks boxes. The table
-        of equivalents usually contains repeated information, so we dump it in compressed form: The
-        sequence of $n + 2$ values $(n, x_1, \ldots, x_n, m)$ in the format file represents $n+m$
-        consecutive entries of |eqtb|, with |m| extra copies of $x_n$, namely $(x_1, \ldots, x_n,
-        x_n, \ldots, x_n)$.
-    */
     int index = null_cs;
+    int n_of_undefined = 0;
+    int n_of_defined = 0;
     do {
-        int different = 1;
-        int equivalent = 0;
+        unsigned char undefined = 0;
+        unsigned char defined = 0;
+# if (pack_them)
+        while (index < first_register_base && undefined < 0xE) {
+# else
+        while (index < first_register_base && undefined < 0xFE) {
+# endif
+             if (equal_eqtb_entries(index, undefined_control_sequence)) {
+                ++undefined;
+                ++index;
+            } else {
+                break;
+            }
+        }
+# if (pack_them)
+        while (index < first_register_base && defined < 0xF) {
+# else
+        while (index < first_register_base && defined < 0xFE) {
+# endif
+             if (! equal_eqtb_entries(index, undefined_control_sequence)) {
+                ++defined;
+                ++index;
+            } else {
+                break;
+            }
+        }
+# if (pack_them)
+        dump_via_uchar(f, (undefined << 4) + defined);
+# else
+        dump_uchar(f, undefined);
+        dump_uchar(f, defined);
+# endif
+        if (defined) {
+            dump_things(f, lmt_hash_state.eqtb[index-defined], defined);
+        }
+        n_of_defined += defined;
+        n_of_undefined += undefined;
+    } while (index < first_register_base);
+# if (pack_them)
+    dump_via_uchar(f, 0xFF);
+# else
+    dump_via_uchar(f, 0xFF);
+    dump_via_uchar(f, 0xFF);
+# endif
+    dump_int(f, n_of_undefined);
+    dump_int(f, n_of_defined);
+}
+
+static void tex_dump_equivalents_mem_registers(dumpstream f) /* the old packer */
+{
+    int index = first_register_base;
+    int n_of_different = 0;
+    int n_of_equivalent = 0;
+    do {
+        unsigned char different = 1;
+        unsigned short equivalent = 0;
         int j = index;
-        while (j < eqtb_size - 1) {
+        while (j < (eqtb_size - 1) && different < 0xFF) {
             if (equal_eqtb_entries(j, j + 1)) {
                 ++equivalent;
                 goto FOUND1;
@@ -217,11 +280,10 @@ void tex_dump_equivalents_mem(dumpstream f)
             }
             ++j;
         }
-        /*tex |j = int_base-1| */
         goto DONE1;
       FOUND1:
         j++;
-        while (j < eqtb_size - 1) {
+        while (j < (eqtb_size - 1) && equivalent < 0xFFFF) {
             if (equal_eqtb_entries(j, j + 1)) {
                 ++equivalent;
             } else {
@@ -230,36 +292,102 @@ void tex_dump_equivalents_mem(dumpstream f)
             ++j;
         }
       DONE1:
-     // printf("index %i, different %i, equivalent %i\n",index,different,equivalent);
-        dump_int(f, different);
-        dump_things(f, lmt_hash_state.eqtb[index], different);
-        dump_int(f, equivalent);
+        dump_uchar(f, different);
+        dump_ushort(f, equivalent);
+        if (different) {
+            dump_things(f, lmt_hash_state.eqtb[index], different);
+        }
         index = index + different + equivalent;
+        n_of_different += different;
+        n_of_equivalent += equivalent;
     } while (index <= eqtb_size);
-    /*tex Dump the |hash_extra| part: */
+    dump_int(f, n_of_different);
+    dump_int(f, n_of_equivalent);
+}
+
+static void tex_dump_equivalents_mem_extra(dumpstream f)
+{
     dump_int(f, lmt_hash_state.hash_data.ptr);
     if (lmt_hash_state.hash_data.ptr > 0) {
+        /*tex Some 15% of the total saved equivalent section (in \CONTEXT\ about 85K). */
         dump_things(f, lmt_hash_state.eqtb[eqtb_size + 1], lmt_hash_state.hash_data.ptr);
     }
-    /*tex A special register. */
+}
+
+static void tex_dump_equivalents_mem_specials(dumpstream f)
+{
     dump_int(f, lmt_token_state.par_loc);
  /* dump_int(f, lmt_token_state.line_par_loc); */ /*tex See note in textoken.c|. */
     dump_int(f, lmt_token_state.empty);
 }
 
-void tex_undump_equivalents_mem(dumpstream f)
+void tex_dump_equivalents_mem(dumpstream f)
+{
+    tex_dump_equivalents_mem_hash(f);
+    tex_dump_equivalents_mem_registers(f);
+    tex_dump_equivalents_mem_extra(f);
+    tex_dump_equivalents_mem_specials(f);
+}
+
+static void tex_undump_equivalents_mem_hash(dumpstream f)
 {
-    /*tex Undump regions 1 to 6 of the table of equivalents |eqtb|. */
     int index = null_cs;
+    memoryword undef = lmt_hash_state.eqtb[undefined_control_sequence];
+    int n_of_undefined = 0;
+    int n_of_defined = 0;
+    while (1) {
+        unsigned char undefined = 0;
+        unsigned char defined = 0;
+# if (pack_them)
+        undump_uchar(f, defined);
+        if (defined == 0xFF) {
+# else
+        undump_uchar(f, undefined);
+        undump_uchar(f, defined);
+        if (undefined == 0xFF && defined == 0xFF) {
+# endif
+            break;
+        } else {
+# if (pack_them)
+            undefined = (defined >> 4) & 0xF;
+            defined = defined & 0xF;
+# endif
+            if (undefined) {
+                for (int i = 1; i <= undefined; i++) {
+                    lmt_hash_state.eqtb[index++] = undef;
+                }
+            }
+            if (defined) {
+                undump_things(f, lmt_hash_state.eqtb[index], defined);
+                index += defined;
+            }
+            n_of_defined += defined;
+            n_of_undefined += undefined;
+        }
+    }
+    {
+        int u, d;
+        undump_int(f, u);
+        undump_int(f, d);
+        if (u != n_of_undefined || d != n_of_defined)  {
+            tex_fatal_undump_error("eqtb count mismatch");
+        }
+    }
+}
+
+static void tex_undump_equivalents_mem_registers(dumpstream f) /* the old unpacker */
+{
+    int index = first_register_base;
+    int n_of_different = 0;
+    int n_of_equivalent = 0;
     do {
-        int different;
-        int equivalent;
-        undump_int(f, different);
-        if (different > 0) {
+        unsigned char different;
+        unsigned short equivalent;
+        undump_uchar(f, different);
+        undump_ushort(f, equivalent);
+        if (different) {
             undump_things(f, lmt_hash_state.eqtb[index], different);
         }
-        undump_int(f, equivalent);
-     // printf("index %i, different %i, equivalent %i\n",index,different,equivalent);
         if (equivalent > 0) {
             int last = index + different - 1;
             for (int i = 1; i <= equivalent; i++) {
@@ -267,13 +395,30 @@ void tex_undump_equivalents_mem(dumpstream f)
             }
         }
         index = index + different + equivalent;
+        n_of_different += different;
+        n_of_equivalent += equivalent;
     } while (index <= eqtb_size);
-    /*tex Undump |hash_extra| part. */
+    {
+        int d, e;
+        undump_int(f, d);
+        undump_int(f, e);
+        if (d != n_of_different || e != n_of_equivalent)  {
+            tex_fatal_undump_error("eqtb register mismatch");
+        }
+    }
+}
+
+static void tex_undump_equivalents_mem_extra(dumpstream f)
+{
     undump_int(f, lmt_hash_state.hash_data.ptr);
     if (lmt_hash_state.hash_data.ptr > 0) {
         /* we get a warning on possible overrun here */
         undump_things(f, lmt_hash_state.eqtb[eqtb_size + 1], lmt_hash_state.hash_data.ptr);
     }
+}
+
+static void tex_undump_equivalents_mem_specials(dumpstream f)
+{
     undump_int(f, lmt_token_state.par_loc);
     if (lmt_token_state.par_loc >= hash_base && lmt_token_state.par_loc <= lmt_hash_state.hash_data.top) {
         lmt_token_state.par_token = cs_token_flag + lmt_token_state.par_loc;
@@ -287,7 +432,14 @@ void tex_undump_equivalents_mem(dumpstream f)
  /*     tex_fatal_undump_error("lineparloc"); */
  /* } */
     undump_int(f, lmt_token_state.empty);
-    return;
+}
+
+void tex_undump_equivalents_mem(dumpstream f)
+{
+    tex_undump_equivalents_mem_hash(f);
+    tex_undump_equivalents_mem_registers(f);
+    tex_undump_equivalents_mem_extra(f);
+    tex_undump_equivalents_mem_specials(f);
 }
 
 /*tex
@@ -405,11 +557,11 @@ static void tex_aux_group_trace(int g)
     \stopitemize
 
     Per end 2023 the decision was made to use the two halfwords in the memory word that is used for
-    saving eq values for other purposes too. It meant for instance that instead if every new group 
-    needing 3 stack entries, it now needs just one. Actually there are not that meny mechanism that 
-    need the stack. For instance alignments have their own stack, math needs it for multi argument 
-    mechanism and the most demanding one is box handling. In addition to the two extra |value_2| and 
-    |value_3| fields we also use |options| and |extra| as alias for |level| (these are quarterwords). 
+    saving eq values for other purposes too. It meant for instance that instead if every new group
+    needing 3 stack entries, it now needs just one. Actually there are not that meny mechanism that
+    need the stack. For instance alignments have their own stack, math needs it for multi argument
+    mechanism and the most demanding one is box handling. In addition to the two extra |value_2| and
+    |value_3| fields we also use |options| and |extra| as alias for |level| (these are quarterwords).
 
 */
 
@@ -454,12 +606,12 @@ static void tex_aux_group_warning(void)
     }
 }
 
-/*tex 
-    We store the line number and attribute state in the memory word part for the save entry and  
-    then we need only one slot instead of three. Each slot is 16 bytes so we're also a bit nicer to 
-    caching memory (if it happens at all). Because in practice we don't group that much (400.000 
-    times in the luametatex manual and a similar amount in the primitives manual) there is no real  
-    significant (positive) impact on performance. It's all about abstraction. 
+/*tex
+    We store the line number and attribute state in the memory word part for the save entry and
+    then we need only one slot instead of three. Each slot is 16 bytes so we're also a bit nicer to
+    caching memory (if it happens at all). Because in practice we don't group that much (400.000
+    times in the luametatex manual and a similar amount in the primitives manual) there is no real
+    significant (positive) impact on performance. It's all about abstraction.
 */
 
 typedef enum saved_group_entries {
@@ -475,7 +627,7 @@ typedef enum saved_group_entries {
 # define saved_group_attribute_state saved_value_2(saved_group_attribute_state_entry)
 # define saved_group_input_line      saved_value_3(saved_group_input_line_entry)
 
-inline static void saved_group_initialize(void)
+static inline void saved_group_initialize(void)
 {
     saved_type(0) = level_boundary_save_type;
     /* here level is really level */
@@ -485,7 +637,7 @@ void tex_new_save_level(quarterword group)
 {
     /*tex We begin a new level of grouping. This will cost us three entries. */
     if (tex_room_on_save_stack()) {
-        add_attribute_reference(current_attribute_state); 
+        add_attribute_reference(current_attribute_state);
         saved_group_initialize();
         saved_group_group = cur_group;
         saved_group_boundary = cur_boundary;
@@ -599,7 +751,7 @@ void tex_show_save_groups(void)
             case output_group:
                 tex_print_str_esc("output");
                 goto FOUND2;
-         // maybe: 
+         // maybe:
          //
          // case math_group:
          // case math_component_group:
@@ -664,16 +816,16 @@ void tex_show_save_groups(void)
                 tex_confusion("show groups");
                 break;
         }
-        /*tex 
-            Show the box context. In traditional \TEX\ the shift is encoded in the context which is 
-            why it had such a large offset for the other context value. That somewhat dirty trick 
+        /*tex
+            Show the box context. In traditional \TEX\ the shift is encoded in the context which is
+            why it had such a large offset for the other context value. That somewhat dirty trick
             was has stepwise been removed.
         */
         switch (tex_get_packaging_context()) {
             case direct_box_flag:
                 {
                     scaled shift = tex_get_packaging_shift();
-                    if (shift != null_flag) { 
+                    if (shift != null_flag) {
                         /*tex We passed the safeguard. */
                         singleword cmd = is_v_mode(lmt_nest_state.nest[pointer].mode) ? hmove_cmd : vmove_cmd;
                         tex_print_cmd_chr(cmd, (shift > 0) ? move_forward_code : move_backward_code);
@@ -738,82 +890,82 @@ void tex_show_save_stack(void)
                 case box_save_type:
                     if (tex_show_packaging_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
                 case local_box_save_type:
                     if (tex_show_localbox_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
                 case alignment_save_type:
                     if (tex_show_alignment_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
                 case adjust_save_type:
                     if (tex_show_adjust_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
                 case math_save_type:
                     if (tex_show_math_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
                 case fraction_save_type:
                     if (tex_show_math_fraction_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
                 case radical_save_type:
                     if (tex_show_math_radical_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
                 case operator_save_type:
                     if (tex_show_math_operator_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
                 case math_group_save_type:
                     if (tex_show_math_group_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
                 case choice_save_type:
                     if (tex_show_math_choice_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
                 case number_save_type:
                     if (tex_show_math_number_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
                 case insert_save_type:
                     if (tex_show_insert_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
                 case discretionary_save_type:
                     if (tex_show_discretionary_record()) {
                         break;
-                    } else { 
+                    } else {
                         goto INVALID_TYPE;
                     }
-                default: 
+                default:
                     goto INVALID_RECORD;
             }
         } else {
@@ -956,7 +1108,7 @@ static int tex_aux_mutation_permitted(halfword cs)
 // static void tex_aux_eq_destroy(memoryword w)
 // {
 //     halfword p = eq_value_field(w);
-//     if (p) { 
+//     if (p) {
 //         switch (eq_type_field(w)) {
 //             case call_cmd:
 //             case protected_call_cmd:
@@ -995,18 +1147,18 @@ static int tex_aux_mutation_permitted(halfword cs)
 //     return eq_value_field(w) ? lmt_hash_state.destructors[eq_type_field(w)] : 0;
 // }
 
-inline static void tex_aux_eq_destroy(memoryword w)
+static inline void tex_aux_eq_destroy(memoryword w)
 {
     halfword p = eq_value_field(w);
-    if (p) { 
+    if (p) {
         switch (lmt_hash_state.destructors[eq_type_field(w)]) {
-            case eq_token_list: 
+            case eq_token_list:
                 tex_delete_token_reference(p);
                 break;
-            case eq_node: 
+            case eq_node:
                 tex_flush_node(p);
                 break;
-            case eq_node_list: 
+            case eq_node_list:
                 tex_flush_node_list(p);
                 break;
         }
@@ -1050,11 +1202,11 @@ static void tex_aux_eq_save(halfword p, quarterword l)
     the case of glue refs that have work by ref count and token lists and node (lists) are always
     different so there we do no harm.
 
-    There is room for some optimization here. 
+    There is room for some optimization here.
 
 */
 
-inline static int tex_aux_equal_eq(halfword p, singleword cmd, singleword flag, halfword chr)
+static inline int tex_aux_equal_eq(halfword p, singleword cmd, singleword flag, halfword chr)
 {
     /* maybe keep flag test at call end and then only flip flags */
     if (eq_flag(p) == flag) {
@@ -1095,6 +1247,14 @@ inline static int tex_aux_equal_eq(halfword p, singleword cmd, singleword flag,
                 } else {
                     return 0;
                 }
+            /*
+            case specificationspec_cmd:
+                if (eq_type(p) == cmd && eq_value(p) == chr) {
+                    return 1;
+                } else {
+                    return 0;
+                }
+            */
             case call_cmd:
             case protected_call_cmd:
             case semi_protected_call_cmd:
@@ -1114,7 +1274,11 @@ inline static int tex_aux_equal_eq(halfword p, singleword cmd, singleword flag,
             case unit_reference_cmd:
             case internal_box_reference_cmd:
             case register_box_reference_cmd:
-                /*tex These are also references. */
+                /*tex These are also references. The ! chr is a bit strange. Todo: test without. */
+// if (eq_type(p) == cmd && eq_value(p) == chr && chr) {
+//     printf("SAME\n");
+//     // destoy ??
+// }
                 if (eq_type(p) == cmd && eq_value(p) == chr && ! chr) {
              // if (eq_type(p) == cmd && eq_value(p) == chr && ! chr && eq_level(p) == cur_level) {
                     return 1;
@@ -1147,7 +1311,7 @@ inline static int tex_aux_equal_eq(halfword p, singleword cmd, singleword flag,
                 if (eq_type(p) == cmd && eq_value(p) == chr) {
              // if (eq_type(p) == cmd && eq_value(p) == chr && eq_level(p) == cur_level) {
                     return 1;
-                } else { 
+                } else {
                     return 0;
                 }
             default:
@@ -1159,7 +1323,7 @@ inline static int tex_aux_equal_eq(halfword p, singleword cmd, singleword flag,
                 if (eq_type(p) == cmd && eq_value(p) == chr) {
              // if (eq_type(p) == cmd && eq_value(p) == chr && eq_level(p) == cur_level) {
                     return 1;
-                } else { 
+                } else {
                     return 0;
                 }
         }
@@ -1270,7 +1434,7 @@ void tex_geq_word_define(halfword p, int w) /* not used */
     a side effect of looking at the code through a visual studio lense.)
 */
 
-inline static void tex_aux_set_eq_data(halfword p, singleword t, halfword e, singleword f, quarterword l)
+static inline void tex_aux_set_eq_data(halfword p, singleword t, halfword e, singleword f, quarterword l)
 {
     singleword flag = eq_flag(p);
     set_eq_level(p, l);
@@ -1321,14 +1485,14 @@ void tex_define(int g, halfword p, singleword t, halfword e) /* int g -> singlew
     }
 }
 
-/*tex 
-    Used in |\dimendef| but also |\dimensiondef| and alike. Before it gets called we already 
-    redefined to |\relax| so we might have saved. 
+/*tex
+    Used in |\dimendef| but also |\dimensiondef| and alike. Before it gets called we already
+    redefined to |\relax| so we might have saved.
 */
 
 void tex_define_again(int g, halfword p, singleword t, halfword e) /* int g -> singleword g */
 {
-    if (tracing_assigns_par > 0) { 
+    if (tracing_assigns_par > 0) {
         singleword f = make_eq_flag_bits(g);
         if (is_global(g)) {
             /* what if already global */
@@ -1345,7 +1509,7 @@ void tex_define_again(int g, halfword p, singleword t, halfword e) /* int g -> s
             tex_aux_diagnostic_trace(p, is_retained(g) ? "retained changing": "changing");
             if (eq_level(p) == cur_level) {
                 tex_aux_eq_destroy(lmt_hash_state.eqtb[p]);
-            } else if (is_retained(g)) { 
+            } else if (is_retained(g)) {
                 /* nothing */
             } else if (cur_level > level_one) {
                 tex_aux_eq_save(p, eq_level(p));
@@ -1353,14 +1517,14 @@ void tex_define_again(int g, halfword p, singleword t, halfword e) /* int g -> s
             tex_aux_set_eq_data(p, t, e, f, cur_level);
         }
         tex_aux_diagnostic_trace(p, "into");
-    } else { 
+    } else {
         set_eq_type(p, t);
-        set_eq_value(p, e);    
+        set_eq_value(p, e);
     }
 }
 
 /*tex
-    The next one is used when we let something. 
+    The next one is used when we let something.
 */
 
 void tex_define_inherit(int g, halfword p, singleword f, singleword t, halfword e)
@@ -1400,8 +1564,8 @@ void tex_define_inherit(int g, halfword p, singleword f, singleword t, halfword
 }
 
 /*tex
-    Used in swapping but beware: when we swap a global vsize with a local ... we can get side 
-    effect. No retain here. 
+    Used in swapping but beware: when we swap a global vsize with a local ... we can get side
+    effect. No retain here.
 */
 
 static void tex_aux_just_define(int g, halfword p, halfword e)
@@ -1451,7 +1615,7 @@ void tex_define_swapped(int g, halfword p1, halfword p2, int force)
            }
         }
         if (v1 == v2)  {
-            return; 
+            return;
         } else {
             switch (t1) {
                 case register_posit_cmd:
@@ -1532,8 +1696,8 @@ void tex_define_swapped(int g, halfword p1, halfword p2, int force)
     );
 }
 
-/*tex 
-    Used in pushing and popping. No retain here. 
+/*tex
+    Used in pushing and popping. No retain here.
 */
 
 void tex_forced_define(int g, halfword p, singleword f, singleword t, halfword e)
@@ -1564,7 +1728,7 @@ void tex_forced_define(int g, halfword p, singleword f, singleword t, halfword e
     }
 }
 
-/*tex 
+/*tex
     Registers and other values that are stored directly without reference.
 */
 
@@ -1700,9 +1864,9 @@ void tex_save_for_after_group(halfword t)
 void tex_unsave(void)
 {
     if (end_of_group_par) {
-        /*tex 
-            This is not yet always ok, and looks like we can get weird commands (in some group 
-            ending situations)! But I need a better example of a failure. (low priority) 
+        /*tex
+            This is not yet always ok, and looks like we can get weird commands (in some group
+            ending situations)! But I need a better example of a failure. (low priority)
         */
         tex_begin_inserted_list(tex_get_available_token(token_val(end_local_cmd, 0)));
         tex_begin_token_list(end_of_group_par, end_of_group_text);
@@ -1710,8 +1874,9 @@ void tex_unsave(void)
             tex_local_control_message("entering token scanner via endgroup");
         }
         tex_local_control(1);
+// tex_cleanup_input_state();
     }
-    delete_attribute_reference(current_attribute_state); 
+    delete_attribute_reference(current_attribute_state);
     tex_unsave_math_codes(cur_level);
     tex_unsave_cat_codes(cat_code_table_par, cur_level);
     tex_unsave_text_codes(cur_level);
@@ -1740,7 +1905,8 @@ void tex_unsave(void)
                             carried out within the group that just ended, the last such definition will
                             therefore survive.
                         */
-                        if (p < internal_integer_base || p > eqtb_size) {
+                     // if (p < internal_integer_base || p > eqtb_size) {
+                        if (p < internal_integer_base || p >= internal_specification_base) {
                             if (eq_level(p) == level_one) {
                                 tex_aux_eq_destroy(save_word(lmt_save_state.save_stack_data.ptr));
                                 if (trace) {
@@ -1830,7 +1996,7 @@ void tex_unsave(void)
         }
         cur_group = saved_group_group;
         cur_boundary = saved_group_boundary;
-        set_current_attribute_state(saved_group_attribute_state); 
+        set_current_attribute_state(saved_group_attribute_state);
     } else {
         /*tex |unsave| is not used when |cur_group=bottom_level| */
         tex_confusion("current level");
@@ -1983,7 +2149,7 @@ void tex_show_cmd_chr(halfword cmd, halfword chr)
     interface to \LUA. We also dropped fonts and some more from the table.
 
     A previous, efficient, still range based variant can be found in the my archive but it makes
-    no sense to keep it commented here (apart from sentimental reasons) so one now only can see 
+    no sense to keep it commented here (apart from sentimental reasons) so one now only can see
     the range agnostic version here.
 
 */
@@ -2149,7 +2315,7 @@ halfword tex_explicit_disc_penalty(halfword mode)
 
 */
 
-inline static void tex_aux_set_eq(halfword base, quarterword level, singleword cmd, halfword value, halfword count)
+static inline void tex_aux_set_eq(halfword base, quarterword level, singleword cmd, halfword value, halfword count)
 {
     if (count > 0) {
         set_eq_level(base, level);
@@ -2196,7 +2362,7 @@ void tex_initialize_equivalents(void)
 
 void tex_initialize_destructors(void)
 {
-    lmt_hash_state.destructors[call_cmd]                         = eq_token_list;                   
+    lmt_hash_state.destructors[call_cmd]                         = eq_token_list;
     lmt_hash_state.destructors[protected_call_cmd]               = eq_token_list;
     lmt_hash_state.destructors[semi_protected_call_cmd]          = eq_token_list;
     lmt_hash_state.destructors[constant_call_cmd]                = eq_token_list;
@@ -2205,7 +2371,7 @@ void tex_initialize_destructors(void)
     lmt_hash_state.destructors[tolerant_semi_protected_call_cmd] = eq_token_list;
     lmt_hash_state.destructors[register_toks_reference_cmd]      = eq_token_list;
     lmt_hash_state.destructors[internal_toks_reference_cmd]      = eq_token_list;
-    lmt_hash_state.destructors[internal_glue_reference_cmd]      = eq_node; 
+    lmt_hash_state.destructors[internal_glue_reference_cmd]      = eq_node;
     lmt_hash_state.destructors[register_glue_reference_cmd]      = eq_node;
     lmt_hash_state.destructors[internal_muglue_reference_cmd]    = eq_node;
     lmt_hash_state.destructors[register_muglue_reference_cmd]    = eq_node;
@@ -2213,9 +2379,10 @@ void tex_initialize_destructors(void)
     lmt_hash_state.destructors[mugluespec_cmd]                   = eq_node;
     lmt_hash_state.destructors[mathspec_cmd]                     = eq_node;
     lmt_hash_state.destructors[fontspec_cmd]                     = eq_node;
+    lmt_hash_state.destructors[specificationspec_cmd]            = eq_node;
     lmt_hash_state.destructors[specification_reference_cmd]      = eq_node;
-    lmt_hash_state.destructors[internal_box_reference_cmd]       = eq_node_list; 
-    lmt_hash_state.destructors[register_box_reference_cmd]       = eq_node_list; 
+    lmt_hash_state.destructors[internal_box_reference_cmd]       = eq_node_list;
+    lmt_hash_state.destructors[register_box_reference_cmd]       = eq_node_list;
 }
 
 int tex_located_save_value(int id)
@@ -2238,7 +2405,7 @@ int tex_located_save_value(int id)
     return 0;
 }
 
-extern int tex_cs_state(halfword p) 
+int tex_cs_state(halfword p)
 {
     if (p == null_cs) {
         return cs_null_error;
@@ -2252,3 +2419,23 @@ extern int tex_cs_state(halfword p)
         return cs_no_error;
     }
 }
+
+void tex_save_stack_catch_up(void)
+{
+    // save_state_info saved_save_stack_data = lmt_save_state;
+    halfword saved_stack_ptr = lmt_save_state.save_stack_data.ptr;
+    quarterword saved_group = cur_group;
+    quarterword saved_level = cur_level;
+    lmt_save_state.save_stack_data.ptr = cur_boundary;
+    while (lmt_input_state.in_stack[lmt_input_state.in_stack_data.ptr].group != lmt_save_state.save_stack_data.ptr) {
+        --cur_level;
+        tex_print_nlp();
+        tex_print_format("Warning: end of file when %G is incomplete", 1);
+        cur_group = save_level(lmt_save_state.save_stack_data.ptr);
+        lmt_save_state.save_stack_data.ptr = save_value(lmt_save_state.save_stack_data.ptr);
+    }
+    // lmt_save_state = saved_save_stack_data;
+    lmt_save_state.save_stack_data.ptr = saved_stack_ptr;
+    cur_level = saved_level;
+    cur_group = saved_group;
+}
