@@ -364,6 +364,7 @@ void tex_print_str(const char *s)
             int newline = s[len-1] == '\n';
             if (logfile) {
                 fputs(s, lmt_print_state.logfile);
+             // fwrite(s, sizeof(char), len, lmt_print_state.logfile);
                 if (newline) {
                     lmt_print_state.logfile_offset = 0;
                 } else {
@@ -372,6 +373,7 @@ void tex_print_str(const char *s)
             }
             if (terminal) {
                 fputs(s, stdout);
+             // fwrite(s, sizeof(char), len, stdout);
                 if (newline) {
                     lmt_print_state.terminal_offset = 0;
                 } else {
@@ -597,6 +599,13 @@ void tex_print_posit(halfword s)
     tex_print_str(b);
 }
 
+void tex_print_posit_5(halfword s)
+{
+    char b[32];
+    sprintf(b, "%.5g", tex_posit_to_double(s));
+    tex_print_str(b);
+}
+
 /*tex
 
     Hexadecimal printing of nonnegative integers is accomplished by |print_hex|. We have a few 
@@ -828,6 +837,7 @@ void tex_print_glue(scaled d, int order, int unit)
 {
     tex_print_dimension(d, no_unit);
     if ((order < normal_glue_order) || (order > filll_glue_order)) {
+        /*tex For sure this will crash someplace becuase it is also used as index! */
         tex_print_str("foul");
     } else if (order > normal_glue_order) {
         tex_print_str("fi");
@@ -961,40 +971,11 @@ void tex_print_char_identifier(halfword c) // todo: use string_print_format
 void tex_print_font_identifier(halfword f)
 {
     /*tex |< >| is less likely to clash with text parenthesis */
+    if (f < 0) { 
+        f = cur_font_par; /* bonus */
+    }
     if (tex_is_valid_font(f)) {
-     // switch (tracing_fonts_par) {
-     //    case 0:
-     //    case 1:
-     //        if (font_original(f)) {
-     //            tex_print_format(font_original(f));
-     //        } else {
-     //            tex_print_format("font: %i", f);
-     //        }
-     //        if (tracing_fonts_par == 0) {
-     //            break;
-     //        } else if (font_size(f) == font_design_size(f)) {
-     //            tex_print_format(" (%s)", font_name(f));
-     //        } else {
-     //            tex_print_format(" (%s @ %D)", font_name(f), font_size(f), pt_unit);
-     //        }
-     //        break;
-     //    case 2:
-     //        tex_print_format("<%s>", font_name(f));
-     //        break;
-     //    case 3:
-     //        tex_print_format("<%s @ %D>", font_name(f), font_size(f), pt_unit);
-     //        break;
-     //    case 4:
-     //        tex_print_format("<%i>", f);
-     //        break;
-     //    case 5:
-     //        tex_print_format("<%i: %s>", f, font_name(f));
-     //        break;
-     // /* case 6: */
-     //    default:
-                tex_print_format("<%i: %s @ %p>", f, font_name(f), font_size(f));
-     //         break;
-     // }
+        tex_print_format("<%i: %s @ %p>", f, font_name(f), font_size(f));
     } else {
         tex_print_str("<*>");
     }
@@ -1011,6 +992,9 @@ void tex_print_font_specifier(halfword e)
 
 void tex_print_font(halfword f)
 {
+    if (f < 0) { 
+        f = cur_font_par; /* bonus */
+    }
     if (! f) {
         tex_print_str("nullfont");
     } else if (tex_is_valid_font(f)) {
@@ -1031,11 +1015,11 @@ void tex_print_font(halfword f)
 
 void tex_short_display(halfword p)
 {
-    tex_print_levels();
+ // tex_print_levels();
     if (p) {
         tex_print_short_node_contents(p);
     } else {
-        tex_print_str("empty list");
+        tex_print_str("[empty list]");
     }
 }
 
